// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file FiniteElemSpace.h */

/** \defgroup FEMSpace FEMSpace
 * @{ <img src="femspace.png"> @}
 */

#ifndef AMDIS_FINITEELEMSPACE_H
#define AMDIS_FINITEELEMSPACE_H

#include <string>
#include <vector>
#include "AMDiS_fwd.h"

namespace AMDiS {

  /** \ingroup FEMSpace
   * \brief
   * A FiniteElemSpace is a triple of a DOFAdmin and a set of BasisFunction on a mesh.
   */
  class FiniteElemSpace
  {
  public:
    /// Create an empty fe space.
    FiniteElemSpace();

    ///
    static FiniteElemSpace *provideFeSpace(DOFAdmin *admin,
					   const BasisFunction *basFcts,
					   Mesh *mesh,
					   std::string name = "");

    /// Destructor.
    ~FiniteElemSpace();
    
    FiniteElemSpace& operator=(const FiniteElemSpace& feSpace);

    /// Returns \ref name.
    inline std::string getName() const 
    { 
      return name;
    }

    /// Returns \ref admin.
    inline DOFAdmin* getAdmin() const 
    { 
      return admin;
    }

    /// Set a new DOF admin.
    inline void setAdmin(DOFAdmin* a) 
    {
      admin = a;
    }

    /// Returns \ref basFcts
    inline const BasisFunction* getBasisFcts() const 
    { 
      return basFcts;
    }

    /// Returns \ref mesh
    inline Mesh* getMesh() const 
    { 
      return mesh; 
    }

    int calcMemoryUsage();

    static void clear();
  
  protected:
    /** \brief
     * Constructs a FiniteElemSpace with name name_ and the given DOFAdmin,
     * BasisFunction and Mesh.
     */
    FiniteElemSpace(DOFAdmin* admin,
		    const BasisFunction* basisFcts, 
		    Mesh* mesh,
		    std::string name = "");

  protected:
    /// Name of this FiniteElemSpace.
    const std::string name;

    /// DOFAdmin corresponding to this FiniteElemSpace.
    DOFAdmin* admin;

    /// Set of BasisFunction of this FiniteElemSpace.
    const BasisFunction* basFcts;

    /// The Mesh this FiniteElemSpace belongs to.
    Mesh* mesh;

    ///
    static std::vector<FiniteElemSpace*> feSpaces;
  };

}

#endif  // !_FINITEELEMSPACE_H_ 
